#include <stdint.h>
#include "../pdl.h"

#define PRECISION 8
#define TO_FIXED(num) (int)((num)*(1<<PRECISION))
#define TO_INT(num) ((num)>>PRECISION)
#define MUL(a, b) ((a*b)>>PRECISION)
#define DIV(a, b) ((a<<PRECISION)/b)

typedef int fixed_t;

#define SGN(a) ((a) < 0 ? -1 : (a) > 0 ? 1 : 0)
#define C_RGB(r,g,b) (((r) << 11) | ((g) << 6) | (b))

#define NB_POINT_CUBE 8
#define NB_LINES_CUBE 12

uint8_t Screen256[ 396 * 224 ];
uint16_t Palette256[ 256 ];

fixed_t scale, eye, eyescale;
int centerx, centery;

fixed_t angle = 0;

typedef struct
{
  fixed_t x;
  fixed_t y;
  fixed_t z;
  fixed_t rx;
  fixed_t ry;
  fixed_t rz;
  int16_t sx;
  int16_t sy;
  int8_t color;
} Vertex3D;

typedef struct
{
    uint8_t p1;
    uint8_t p2;
} Pairs;

Pairs Lines[NB_LINES_CUBE];
Vertex3D Cube[NB_POINT_CUBE];

/* See
 * https://en.wikipedia.org/wiki/Bh%C4%81skara_I%27s_sine_approximation_formula.
 */
fixed_t sine(fixed_t d) {
    fixed_t v;
    fixed_t org_d = d%TO_FIXED(360);
    while(org_d < TO_FIXED(0)) org_d += TO_FIXED(360);
    d = org_d;
    d %= TO_FIXED(180);
    v = DIV(MUL(MUL(TO_FIXED(4), d), (TO_FIXED(180)-d)),
            (TO_FIXED(40500)-MUL(d, (TO_FIXED(180)-d))));
    if(org_d > TO_FIXED(180)) v = -v;
    return v;
}

/* See
 * https://en.wikipedia.org/wiki/Bh%C4%81skara_I%27s_sine_approximation_formula.
 */
fixed_t cosine(fixed_t d) {
    return sine( d + TO_FIXED(90) );
}

void ProjectVertex( Vertex3D *vertex )
{
	fixed_t divisor = DIV( eyescale, (MUL( scale, vertex->rz ) + eye) );

    vertex->sx = centerx + TO_INT( MUL( vertex->rx, divisor ) );
	vertex->sy = centery + TO_INT( MUL( vertex->ry, divisor ) );
}

void RotateVertex( Vertex3D *vertex, fixed_t ax, fixed_t ay, fixed_t az )
{
	// Rotate around x axis
	vertex->ry = MUL(vertex->y, cosine(ax)) - MUL(vertex->z , sine(ax));
	vertex->rz = MUL(vertex->y, sine(ax)) + MUL(vertex->z, cosine(ax));

	// Rotate around y axis
	vertex->rx = MUL(vertex->x, cosine(ay)) + MUL(vertex->rz, sine(ay));
	vertex->rz =  MUL(vertex->rz, cosine(ay)) - MUL(vertex->x, sine(ay));

	// Rotate around z axis
	fixed_t tmpx = MUL(vertex->rx, cosine(az)) - MUL(vertex->ry, sine(az));
	vertex->ry = MUL(vertex->rx, sine(az)) + MUL(vertex->ry, cosine(az));
	vertex->rx = tmpx;
}

void pixel( uint8_t *buffer, uint16_t sx, uint16_t sy, uint8_t color )
{
	if (sx>=396 || sy>=224 ) return;
	buffer[ sx + sy * 396 ] = color;
}

void Drawline( uint8_t *buffer, int x1, int y1, int x2, int y2, uint8_t color)
{
	int i;
    int x = x1;
    int y = y1;
    int cumul;
	int dx = x2 - x1;
    int dy = y2 - y1;
	int sx = SGN(dx);
    int sy = SGN(dy);

	dx = (dx >= 0 ? dx : -dx);
    dy = (dy >= 0 ? dy : -dy);

	pixel(buffer, x1, y1, color);

	if(dx >= dy)
	{
		cumul = dx >> 1;
		for(i = 1; i < dx; i++)
		{
			x += sx;
			cumul += dy;
			if(cumul > dx) cumul -= dx, y += sy;
			pixel(buffer, x, y, color);
		}
	}
	else
	{
		cumul = dy >> 1;
		for(i = 1; i < dy; i++)
		{
			y += sy;
			cumul += dx;
			if(cumul > dy) cumul -= dy, x += sx;
			pixel(buffer, x, y, color);
		}
	}

	pixel(buffer, x2, y2, color);
}

void Blur( uint8_t *buffer )
{
	uint16_t temp;
    uint32_t offset = 223 * 396;                    // pixel(x,y)
    
    for( int index = 0; index < 223; index++ )
    {
        for( int x = 1;  x < 396 - 1; x++ )
        {
            temp = buffer[ offset + x - 1 ] +       // pixel(x-1,y)
                   buffer[ offset + x ] +           // pixel(x,y)
                   buffer[ offset + x + 1 ] +       // pixel(x+1,y)
                   buffer[ offset + x - 396 ];      // pixel(x,y-1)
            buffer[offset + x - 396] = (uint8_t) (temp >> 2);
        }
        offset -= 396;
    }
}

void Dclear( uint8_t *Screen, uint8_t color )
{
	for(int i = 0; i < 396 * 224; i++)
		Screen[i] = color;
}

void Render( uint16_t *vram, uint8_t *Screen, uint16_t *Palette )
{
	for(int i = 0; i < 396 * 224; i++)
		vram[i] = Palette[ Screen[i] ];
}

void SetPalette( uint16_t *Palette )
{
	for(int u = 0; u<64; u++)
    {
        Palette[u]     = C_RGB(   0,   0,   0 );
        Palette[u+64]  = C_RGB( u/2,   0,   0 );
        Palette[u+128] = C_RGB(  31, u/2,   0 );
        Palette[u+192] = C_RGB(  31,  31, u/2 );
    }
}

void FirecubeInit( void )
{
    eye =  TO_FIXED(250);
    scale =  TO_FIXED(60);
    eyescale = MUL(eye, scale);

    centerx = (396 / 2);
    centery = (224 / 2);
    
    Cube[0].x = TO_FIXED(-0.75);    Cube[0].y = TO_FIXED(-0.75);    Cube[0].z = TO_FIXED(-0.75); 
    Cube[1].x = TO_FIXED(0.75);     Cube[1].y = TO_FIXED(-0.75);    Cube[1].z = TO_FIXED(-0.75); 
    Cube[2].x = TO_FIXED(0.75);     Cube[2].y = TO_FIXED(0.75);     Cube[2].z = TO_FIXED(-0.75); 
    Cube[3].x = TO_FIXED(-0.75);    Cube[3].y = TO_FIXED(0.75);     Cube[3].z = TO_FIXED(-0.75); 

    Cube[4].x = TO_FIXED(-0.75);    Cube[4].y = TO_FIXED(-0.75);    Cube[4].z = TO_FIXED(0.75); 
    Cube[5].x = TO_FIXED(0.75);     Cube[5].y = TO_FIXED(-0.75);    Cube[5].z = TO_FIXED(0.75); 
    Cube[6].x = TO_FIXED(0.75);     Cube[6].y = TO_FIXED(0.75);     Cube[6].z = TO_FIXED(0.75); 
    Cube[7].x = TO_FIXED(-0.75);    Cube[7].y = TO_FIXED(0.75);     Cube[7].z = TO_FIXED(0.75); 

    Lines[0].p1 = 0; Lines[0].p2 = 1;
    Lines[1].p1 = 1; Lines[1].p2 = 2;
    Lines[2].p1 = 2; Lines[2].p2 = 3;
    Lines[3].p1 = 3; Lines[3].p2 = 0;

    Lines[4].p1 = 4; Lines[4].p2 = 5;
    Lines[5].p1 = 5; Lines[5].p2 = 6;
    Lines[6].p1 = 6; Lines[6].p2 = 7;
    Lines[7].p1 = 7; Lines[7].p2 = 4;

    Lines[8].p1 = 0; Lines[8].p2 = 4;
    Lines[9].p1 = 1; Lines[9].p2 = 5;
    Lines[10].p1 = 2; Lines[10].p2 = 6;
    Lines[11].p1 = 3; Lines[11].p2 = 7;
}

void FirecubeUpdate( void )
{
	angle += 1;
    for (int i = 0; i < NB_POINT_CUBE; i++)
    {
		RotateVertex(&Cube[i], 0, TO_FIXED(angle), TO_FIXED(angle));
		ProjectVertex(&Cube[i]);
	}
}

void FirecubeRender( uint8_t *screen )
{ 
    for (int j = 0; j < NB_LINES_CUBE; j++)
        Drawline( screen, Cube[ Lines[j].p1 ].sx, Cube[ Lines[j].p1 ].sy, Cube[ Lines[j].p2 ].sx, Cube[ Lines[j].p2 ].sy, 255 );
    Blur( screen );
}


PDL_MAIN int main(void)
{
	SetPalette( Palette256 );
	Dclear( Screen256, 0 );

	FirecubeInit( );

	while(PDL->time_ms <= 30000)
	{
		FirecubeUpdate( );
		FirecubeRender( Screen256 );
		Render( PDL->vram, Screen256, Palette256 );

		PDL->dupdate();
	}

	return 0;
}
